#/usr/bin/python

import logging
import itertools
import os.path
import phonenumbers
import phonenumbers.geocoder
import tornado.locale
import tornado.options
import tornado.web

import ipfire
import ipfire.countries

from .handlers import *

from . import auth
from . import blog
from . import boot
from . import donate
from . import download
from . import fireinfo
from . import iuse
from . import location
from . import mirrors
from . import nopaste
from . import people
from . import ui_modules
from . import wiki

class Application(tornado.web.Application):
	def __init__(self, config, **kwargs):
		# Initialize backend
		self.backend = ipfire.Backend(config)

		settings = {
			# Do not compress responses
			"gzip" : False,

			# Enable XSRF cookies
			"xsrf_cookies" : True,

			# Login
			"login_url" : "/login",

			# Setup directory structure
			"static_path"   : self.backend.config.get("global", "static_dir"),
			"template_path" : self.backend.config.get("global", "templates_dir"),

			# UI Methods
			"ui_methods" : {
				"format_country_name"          : self.format_country_name,
				"format_language_name"         : self.format_language_name,
				"format_month_name"            : self.format_month_name,
				"format_phone_number"          : self.format_phone_number,
				"format_phone_number_to_e164"  : self.format_phone_number_to_e164,
				"format_phone_number_location" : self.format_phone_number_location,
				"grouper"                      : grouper,
			},

			# UI Modules
			"ui_modules" : {
				# Blog
				"BlogHistoryNavigation": blog.HistoryNavigationModule,
				"BlogList"             : blog.ListModule,
				"BlogPost"             : blog.PostModule,
				"BlogPosts"            : blog.PostsModule,

				# Boot
				"BootMenuConfig"       : boot.MenuConfigModule,
				"BootMenuHeader"       : boot.MenuHeaderModule,
				"BootMenuSeparator"    : boot.MenuSeparatorModule,

				# People
				"AccountsList"         : people.AccountsListModule,
				"Agent"                : people.AgentModule,
				"CDR"                  : people.CDRModule,
				"Channels"             : people.ChannelsModule,
				"MOS"                  : people.MOSModule,
				"NewAccounts"          : people.NewAccountsModule,
				"Password"             : people.PasswordModule,
				"Registrations"        : people.RegistrationsModule,
				"SIPStatus"            : people.SIPStatusModule,

				# Nopaste
				"Code"                 : nopaste.CodeModule,

				# Fireinfo
				"FireinfoDeviceTable"  : fireinfo.DeviceTableModule,
				"FireinfoDeviceAndGroupsTable"
				                       : fireinfo.DeviceAndGroupsTableModule,

				# Wiki
				"WikiDiff"             : wiki.WikiDiffModule,
				"WikiNavbar"           : wiki.WikiNavbarModule,
				"WikiList"             : wiki.WikiListModule,

				# Misc
				"ChristmasBanner"      : ui_modules.ChristmasBannerModule,
				"Markdown"             : ui_modules.MarkdownModule,
				"Map"                  : ui_modules.MapModule,
				"ProgressBar"          : ui_modules.ProgressBarModule,
			},

			# Call this when a page wasn't found
			"default_handler_class" : base.NotFoundHandler,
		}
		settings.update(kwargs)

		tornado.web.Application.__init__(self, **settings)

		authentication_handlers = [
			(r"/login", auth.LoginHandler),
			(r"/logout", auth.LogoutHandler),
		]

		self.add_handlers(r"(dev|www)\.ipfire\.org", [
			# Entry site that lead the user to index
			(r"/", IndexHandler),

			# Download sites
			(r"/downloads", tornado.web.RedirectHandler, { "url" : "/download" }),
			(r"/download", download.IndexHandler),
			(r"/download/([0-9a-z\-\.]+)", download.ReleaseHandler),

			# Donate
			(r"/donate", donate.DonateHandler),
			(r"/donate/thank-you", donate.ThankYouHandler),
			(r"/donate/error", donate.ErrorHandler),
			(r"/donation", tornado.web.RedirectHandler, { "url" : "/donate" }),

			# RSS feed
			(r"/news.rss", tornado.web.RedirectHandler, { "url" : "https://blog.ipfire.org/feed.xml" }),

			# Redirect news articles to blog
			(r"/news/(.*)", handlers.NewsHandler),

			# Static Pages
			(r"/features", StaticHandler, { "template" : "features.html" }),
			(r"/legal", StaticHandler, { "template" : "legal.html" }),
			(r"/support", StaticHandler, { "template" : "support.html" }),

			# Handle old pages that have moved elsewhere
			(r"/imprint", tornado.web.RedirectHandler, { "url" : "/legal" }),
			(r"/(de|en)/(.*)", LangCompatHandler),

			# Export arbitrary error pages
			(r"/error/([45][0-9]{2})", base.ErrorHandler),

			# Block page
			(r"/blocked", base.BlockedHandler),
		])

		# blog.ipfire.org
		self.add_handlers(r"blog(\.dev)?\.ipfire\.org", [
			(r"/", blog.IndexHandler),
			(r"/authors/(\w+)", blog.AuthorHandler),
			(r"/compose", blog.ComposeHandler),
			(r"/drafts", blog.DraftsHandler),
			(r"/post/([0-9a-z\-\._]+)", blog.PostHandler),
			(r"/post/([0-9a-z\-\._]+)/delete", blog.DeleteHandler),
			(r"/post/([0-9a-z\-\._]+)/edit", blog.EditHandler),
			(r"/post/([0-9a-z\-\._]+)/publish", blog.PublishHandler),
			(r"/search", blog.SearchHandler),
			(r"/tags/([0-9a-z\-\.]+)", blog.TagHandler),
			(r"/years/([0-9]+)", blog.YearHandler),

			# RSS Feed
			(r"/feed.xml", blog.FeedHandler),
		] + authentication_handlers)

		# downloads.ipfire.org
		self.add_handlers(r"downloads?(\.dev)?\.ipfire\.org", [
			(r"/", tornado.web.RedirectHandler, { "url" : "https://www.ipfire.org/" }),
			(r"/release/(.*)", download.ReleaseRedirectHandler),
			(r"/(.*)", download.FileHandler),
		])

		# mirrors.ipfire.org
		self.add_handlers(r"mirrors(\.dev)?\.ipfire\.org", [
			(r"/", mirrors.IndexHandler),
			(r"/mirrors/(.*)", mirrors.MirrorHandler),
		])

		# planet.ipfire.org
		self.add_handlers(r"planet(\.dev)?\.ipfire\.org", [
			(r"/", tornado.web.RedirectHandler, { "url" : "https://blog.ipfire.org/" }),
			(r"/post/([A-Za-z0-9_-]+)", handlers.PlanetPostHandler),
			(r"/user/([a-z0-9_-]+)", handlers.PlanetUserHandler),

			# RSS
			(r"/rss", tornado.web.RedirectHandler, { "url" : "https://blog.ipfire.org/feed.xml" }),
			(r"/user/([a-z0-9_-]+)/rss", tornado.web.RedirectHandler, { "url" : "https://blog.ipfire.org/feed.xml" }),
			(r"/news.rss", tornado.web.RedirectHandler, { "url" : "https://blog.ipfire.org/feed.xml" }),
		])

		# fireinfo.ipfire.org
		self.add_handlers(r"fireinfo(\.dev)?\.ipfire\.org", [
			(r"/", fireinfo.IndexHandler),

			# Vendors
			(r"/vendors", fireinfo.VendorsHandler),
			(r"/vendors/(pci|usb)/([0-9a-f]{4})", fireinfo.VendorHandler),

			# Driver
			(r"/drivers/(.*)", fireinfo.DriverDetail),

			# Show profiles
			(r"/profile/random", fireinfo.RandomProfileHandler),
			(r"/profile/([a-z0-9]{40})", fireinfo.ProfileHandler),

			# Stats
			(r"/processors", fireinfo.ProcessorsHandler),
			(r"/releases", fireinfo.ReleasesHandler),

			# Send profiles
			(r"/send/([a-z0-9]+)", fireinfo.ProfileSendHandler),
		])

		# i-use.ipfire.org
		self.add_handlers(r"i-use(\.dev)?\.ipfire\.org", [
			(r"/", tornado.web.RedirectHandler, { "url" : "https://www.ipfire.org/" }),
			(r"/profile/([a-f0-9]{40})/([0-9]+).png", iuse.ImageHandler),
		])

		# boot.ipfire.org
		BOOT_STATIC_PATH = os.path.join(self.settings["static_path"], "netboot")
		self.add_handlers(r"boot(\.dev)?\.ipfire\.org", [
			(r"/", tornado.web.RedirectHandler, { "url" : "https://wiki.ipfire.org/installation/pxe" }),

			# Configurations
			(r"/premenu.cfg", boot.PremenuCfgHandler),
			(r"/menu.gpxe", boot.MenuGPXEHandler),
			(r"/menu.cfg", boot.MenuCfgHandler),

			# Static files
			(r"/(boot\.png|pxelinux\.0|menu\.c32|vesamenu\.c32)",
				tornado.web.StaticFileHandler, { "path" : BOOT_STATIC_PATH }),
		])

		# nopaste.ipfire.org
		self.add_handlers(r"nopaste(\.dev)?\.ipfire\.org", [
			(r"/", nopaste.CreateHandler),
			(r"/raw/(.*)", nopaste.RawHandler),
			(r"/view/(.*)", nopaste.ViewHandler),
		] + authentication_handlers)

		# location.ipfire.org
		self.add_handlers(r"location(\.dev)?\.ipfire\.org", [
			(r"/", location.IndexHandler),
			(r"/how\-to\-use", StaticHandler, { "template" : "../location/how-to-use.html" }),
			(r"/lookup/(.+)/blacklists", location.BlacklistsHandler),
			(r"/lookup/(.+)", location.LookupHandler),
		])

		# geoip.ipfire.org
		self.add_handlers(r"geoip(\.dev)?\.ipfire\.org", [
			(r"/", tornado.web.RedirectHandler, { "url" : "https://location.ipfire.org/" }),
		])

		# talk.ipfire.org
		self.add_handlers(r"talk(\.dev)?\.ipfire\.org", [
			(r"/", tornado.web.RedirectHandler, { "url" : "https://people.ipfire.org/" }),
		])

		# people.ipfire.org
		self.add_handlers(r"people(\.dev)?\.ipfire\.org", [
			(r"/", people.IndexHandler),
			(r"/activate/([a-z_][a-z0-9_-]{0,31})/(\w+)", auth.ActivateHandler),
			(r"/conferences", people.ConferencesHandler),
			(r"/groups", people.GroupsHandler),
			(r"/groups/([a-z_][a-z0-9_-]{0,31})", people.GroupHandler),
			(r"/register", auth.RegisterHandler),
			(r"/search", people.SearchHandler),
			(r"/users", people.UsersHandler),
			(r"/users/([a-z_][a-z0-9_-]{0,31})", people.UserHandler),
			(r"/users/([a-z_][a-z0-9_-]{0,31})\.jpg", people.AvatarHandler),
			(r"/users/([a-z_][a-z0-9_-]{0,31})/calls/([0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12})", people.CallHandler),
			(r"/users/([a-z_][a-z0-9_-]{0,31})/calls(?:/(\d{4}-\d{2}-\d{2}))?", people.CallsHandler),
			(r"/users/([a-z_][a-z0-9_-]{0,31})/edit", people.UserEditHandler),
			(r"/users/([a-z_][a-z0-9_-]{0,31})/passwd", people.UserPasswdHandler),
			(r"/users/([a-z_][a-z0-9_-]{0,31})/sip", people.SIPHandler),

			# Promotional Consent Stuff
			(r"/subscribe", people.SubscribeHandler),
			(r"/unsubscribe", people.UnsubscribeHandler),

			# Single-Sign-On for Discourse
			(r"/sso/discourse", people.SSODiscourse),

			# Password Reset
			(r"/password\-reset", auth.PasswordResetInitiationHandler),
			(r"/password\-reset/([a-z_][a-z0-9_-]{0,31})/(\w+)", auth.PasswordResetHandler),

			# Stats
			(r"/stats", people.StatsHandler),

			# API
			(r"/api/check/uid", auth.APICheckUID),
		]  + authentication_handlers)

		# wiki.ipfire.org
		self.add_handlers(r"wiki(\.dev)?\.ipfire\.org",
			authentication_handlers + [

			# Actions
			(r"((?:[A-Za-z0-9\-_\/]+)?(?:.*)\.(?:\w+))/_delete", wiki.ActionDeleteHandler),
			(r"([A-Za-z0-9\-_\/]+)?/_edit", wiki.ActionEditHandler),
			(r"([A-Za-z0-9\-_\/]+)?/_render", wiki.ActionRenderHandler),
			(r"([A-Za-z0-9\-_\/]+)?/_(watch|unwatch)", wiki.ActionWatchHandler),
			(r"/actions/restore", wiki.ActionRestoreHandler),
			(r"/actions/upload", wiki.ActionUploadHandler),

			# Handlers
			(r"/recent\-changes", wiki.RecentChangesHandler),
			(r"/search", wiki.SearchHandler),
			(r"/watchlist", wiki.WatchlistHandler),

			# Media
			(r"([A-Za-z0-9\-_\/]+)?/_files", wiki.FilesHandler),
			(r"((?!/static)(?:[A-Za-z0-9\-_\/]+)?(?:.*)\.(?:\w+))$", wiki.FileHandler),

			# Render pages
			(r"([A-Za-z0-9\-_\/]+)?", wiki.PageHandler),
		])

		# ipfire.org
		self.add_handlers(r"ipfire\.org", [
			(r".*", tornado.web.RedirectHandler, { "url" : "https://www.ipfire.org" })
		])

		logging.info("Successfully initialied application")

	def format_country_name(self, handler, country_code):
		return ipfire.countries.get_name(country_code)

	def format_language_name(self, handler, language):
		_ = handler.locale.translate

		if language == "de":
			return _("German")
		elif language == "en":
			return _("English")
		elif language == "es":
			return _("Spanish")
		elif language == "fr":
			return _("French")
		elif language == "it":
			return _("Italian")
		elif language == "nl":
			return _("Dutch")
		elif language == "pl":
			return _("Polish")
		elif language == "pt":
			return _("Portuguese")
		elif language == "ru":
			return _("Russian")
		elif language == "tr":
			return _("Turkish")

		return language

	def format_month_name(self, handler, month):
		_ = handler.locale.translate

		if month == 1:
			return _("January")
		elif month == 2:
			return _("February")
		elif month == 3:
			return _("March")
		elif month == 4:
			return _("April")
		elif month == 5:
			return _("May")
		elif month == 6:
			return _("June")
		elif month == 7:
			return _("July")
		elif month == 8:
			return _("August")
		elif month == 9:
			return _("September")
		elif month == 10:
			return _("October")
		elif month == 11:
			return _("November")
		elif month == 12:
			return _("December")

		return month

	def format_phone_number(self, handler, number):
		if not isinstance(number, phonenumbers.PhoneNumber):
			try:
				number = phonenumbers.parse(number, None)
			except phonenumbers.phonenumberutil.NumberParseException:
				return number

		return phonenumbers.format_number(number, phonenumbers.PhoneNumberFormat.INTERNATIONAL)

	def format_phone_number_to_e164(self, handler, number):
		return phonenumbers.format_number(number, phonenumbers.PhoneNumberFormat.E164)

	def format_phone_number_location(self, handler, number):
		s = [
			phonenumbers.geocoder.description_for_number(number, handler.locale.code),
			phonenumbers.region_code_for_number(number),
		]

		return ", ".join((e for e in s if e))


def grouper(handler, iterator, n):
	"""
		Returns groups of n from the iterator
	"""
	i = iter(iterator)

	while True:
		ret = list(itertools.islice(i, 0, n))
		if not ret:
			break

		yield ret
