#!/usr/bin/python3

import iso3166
import tornado.web

from . import base

class DonateHandler(base.BaseHandler):
	def get(self):
		location = self.get_remote_location()

		if location:
			country = location.country
		else:
			country = None

		# Get defaults
		first_name = self.get_argument("first_name", None)
		last_name = self.get_argument("last_name", None)
		amount    = self.get_argument_float("amount", None)
		currency  = self.get_argument("currency", None)
		frequency = self.get_argument("frequency", None)

		# Set default currency
		if not currency in ("EUR", "USD"):
			currency = "EUR"

			# Default to USD for the US only
			if country == "US":
				currency = "USD"

		# Set default frequency
		if not frequency in ("one-time", "monthly"):
			frequency = "one-time"

		self.render("donate/donate.html", countries=iso3166.countries,
			country=country, first_name=first_name, last_name=last_name,
			amount=amount, currency=currency, frequency=frequency)

	@base.ratelimit(minutes=15, requests=5)
	async def post(self):
		amount    = self.get_argument("amount")
		currency  = self.get_argument("currency", "EUR")
		frequency = self.get_argument("frequency")

		# Get form inputs
		args = {
			"amount"       : amount,
			"currency"     : currency,

			# Is this a recurring donation?
			"recurring"    : frequency == "monthly",

			# Address
			"email"        : self.get_argument("email"),
			"title"        : self.get_argument("title"),
			"first_name"   : self.get_argument("first_name"),
			"last_name"    : self.get_argument("last_name"),
			"company_name" : self.get_argument("company_name", None),
			"street1"      : self.get_argument("street1"),
			"street2"      : self.get_argument("street2", None),
			"post_code"    : self.get_argument("post_code"),
			"city"         : self.get_argument("city"),
			"state"        : self.get_argument("state", None),
			"country_code" : self.get_argument("country_code"),
		}

		# Add URLs to redirect the user back
		args.update({
			"success_url"  : "https://%s/donate/thank-you" % self.request.host,
			"error_url"    : "https://%s/donate/error" % self.request.host,
			"back_url"     : "https://%s/donate?amount=%s&currency=%s&frequency=%s" %
				(self.request.host, amount, currency, frequency),
		})

		# Send request to Zeiterfassung
		try:
			response = await self.backend.zeiterfassung.send_request(
				"/api/v1/donations/create/ipfire-project", **args)

		except Exception:
			raise # XXX handle any problems when Zeiterfassung is unreachable

		# Redirect the user to the payment page
		redirect_url = response.get("redirect_url")
		if not redirect_url:
			raise tornado.web.HTTPError(500, "Did not receive a redirect URL")

		self.redirect(redirect_url)


class ThankYouHandler(base.BaseHandler):
    def get(self):
        self.render("donate/thank-you.html")


class ErrorHandler(base.BaseHandler):
    def get(self):
        self.render("donate/error.html")
